"""Module for scene geometry handling and basic algebraic methods."""
try:
    import numba
    prange = numba.prange
except ImportError:
    numba = None
    prange = range
import numpy as np
import matplotlib.axes
import matplotlib.pyplot as plt

from sparrowpy.sound_object import Receiver, SoundSource


class Polygon():
    """Polygon constructed from greater than two points.

    Only convex polygons are allowed!
    Order of points is of course important!
    """

    up_vector: np.ndarray
    pts: np.ndarray

    def __init__(
            self, points: np.ndarray, up_vector: np.ndarray,
            normal: np.ndarray) -> None:
        """Create a Polygon from points, up_vector and normal.

        Parameters
        ----------
        points : np.ndarray
            Cartesian edge points of the polygon
        up_vector : np.ndarray
            Cartesian up vector of the polygon
        normal : np.ndarray
            Cartesian up normal of the polygon

        """
        self.pts = np.array(points, dtype=float)
        normal = np.array(normal, dtype=float)
        self.form_factors = []
        # check if points are in one plane
        assert self.pts.shape[1] >= 3, \
            'You need at least 3 points to build a Polygon'
        if self.n_points > 3:
            x_0 = np.array(self.pts[0])
            for i in range(1, self.n_points-2):
                # the determinant of the vectors (volume) must always be 0
                x_i = np.array(self.pts[i])
                x_i1 = np.array(self.pts[i+1])
                x_i2 = np.array(self.pts[i+2])
                det = np.linalg.det([x_0-x_i, x_0-x_i1, x_0-x_i2])
                assert _cmp_floats(det, 0.0), \
                    'Points must be in a plane to create a Polygon'
        self.up_vector = _norm(np.array(up_vector, dtype=float))
        vec1 = np.array(self.pts[0])-np.array(self.pts[1])
        vec2 = np.array(self.pts[0])-np.array(self.pts[2])
        calc_normal = _norm(np.cross(vec1, vec2))
        assert all(np.abs(np.cross(normal, calc_normal)) < 1e-10), \
            'The normal vector is not perpendicular to the polygon'
        self._normal = normal

    def to_dict(self):
        """Convert this object to dictionary. Used for read write."""
        return {
            'up_vector': self.up_vector.tolist(),
            'pts': self.pts.tolist(),
            'normal': self._normal.tolist(),
        }

    @classmethod
    def from_dict(cls, input_dict):
        """Create an object from a dictionary. Used for read write."""
        return cls(
            input_dict['pts'], input_dict['up_vector'], input_dict['normal'])

    @property
    def normal(self) -> np.ndarray:
        """Return the normal vector of the polygon."""
        return self._normal

    @property
    def size(self) -> np.ndarray:
        """Return the size in (lxmxn) of the polygon."""
        vec1 = np.array(self.pts[0])-np.array(self.pts[1])
        vec2 = np.array(self.pts[1])-np.array(self.pts[2])
        size = np.abs(vec1-vec2)
        return size

    @property
    def area(self) -> np.ndarray:
        """Return the area in m^2 of the polygon.
        supports all convex polygons and some concave polygons.
        """

        area = 0

        if len(self.pts) == 3:
            area_pts = np.array([self.pts])

        elif len(self.pts) == 4:
            area_pts = np.array([
                self.pts[0:3],[self.pts[2],self.pts[3],self.pts[0]] ])

        else:
            # slow, can be optimized
            area_pts = np.empty((self.pts.shape[0],3,3))

            for i in range(area_pts.shape[0]):
                area_pts[i] = np.array([
                    self.pts[i%self.pts.shape[0]],
                    self.pts[(i+1)%self.pts.shape[0]],
                    self.center])

        for tri in area_pts:
            area  +=  .5*np.linalg.norm(np.cross(tri[1]-tri[0], tri[2]-tri[0]))

        return area

    @property
    def center(self) -> np.ndarray:
        """Return the center coordinates of the polygon."""
        return np.sum(self.pts, axis=0) / self.n_points

    @property
    def n_points(self) -> int:
        """Return the number of points of the polygon."""
        return self.pts.shape[0]

    def on_surface(self, point: np.ndarray) -> bool:
        """Return if a point is on the surface of the polygon.

        Returns True if the point is on the polygon's
        surface and false otherwise.
        """
        n = self.pts.shape[0]
        sum_angle = 0
        p = point

        for i in range(n):
            v1 = np.array(self.pts[i]) - p
            v2 = np.array(self.pts[(i+1) % n]) - p

            m1 = _magnitude(v1)
            m2 = _magnitude(v2)

            if _cmp_floats(m1*m2, 0.):
                return True  # point is one of the nodes
            else:
                cos_theta = np.dot(v1, v2)/(m1*m2)
            sum_angle = sum_angle + np.arccos(cos_theta)
        return _cmp_floats(sum_angle, 2*np.pi)

    def intersection(
            self, origin: np.ndarray, direction: np.ndarray) -> np.ndarray:
        """Return a intersection point with a ray and the polygon.

        Parameters
        ----------
        origin : np.ndarray
            origin of the incoming wave
        direction : np.ndarray
            direction of the incoming wave

        Returns
        -------
        np.ndarray
            intersection point, if it hit, otherwise None

        """
        origin = np.asarray(origin, dtype=float)
        direction = np.asarray(direction, dtype=float)
        n = self.normal

        # Ray is parallel to the polygon
        if _cmp_floats(np.dot(direction, n), 0.):
            return None

        t = 1/(np.dot(direction, n)) * \
            (np.dot(n, self.pts[0]) -
             np.dot(n, origin))

        # Intersection point is behind the ray
        if t <= 0.0:
            return None

        # Calculate intersection point
        point = np.array(origin) + t*np.array(direction)

        # Check if intersection point is really in the polygon or only on
        # the (infinite) plane
        if self.on_surface(point):
            return point

        return None
# I changed this part
    def plot(self, ax: matplotlib.axes.Axes = None, color=None):
        """Plot the polygon.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            _description_, by default None
        color : _type_, optional
            _description_, by default None

        """
        points = self.pts.T
        points = np.concatenate((points, points), axis=1)
        # plot wall
        self.plot_point(ax, color)

    def plot_point(self, ax: matplotlib.axes.Axes = None, color=None):
        """Plot the polygon points."""
        points = self.pts.T
        points = np.concatenate((points, points), axis=1)
        # plot wall
        ax.plot(points[0], points[1], points[2], color=color)

    def plot_view_up(self, ax: matplotlib.axes.Axes = None):
        """Plot the view and up vector of the polygon."""
        ax.quiver(
            self.center[0], self.center[1], self.center[2],
            self.normal[0]*2, self.normal[1]*2, self.normal[2]*2,
            color='red', label='View vector')
        ax.quiver(
            self.center[0], self.center[1], self.center[2],
            self.up_vector[0]*2, self.up_vector[1]*2, self.up_vector[2]*2,
            color='blue', label='up vector')

class _Environment():
    """Define a Geometry with Walls, Source and Receiver."""

    speed_of_sound: float
    polygons: list[Polygon]
    source: SoundSource
    receiver: Receiver

    def __init__(
            self, polygons: list[Polygon], source: SoundSource,
            receiver: Receiver, speed_of_sound: float) -> None:
        """Define environment with acoustic Objects and speed of sound.

        Parameters
        ----------
        polygons : list[Polygon]
            input polygons as a list
        source : SoundSource
            sound source in the scene
        receiver : Receiver
            receiver in the scene
        speed_of_sound : float, optional
            speed of sound in m/s

        """
        self.speed_of_sound = speed_of_sound
        self.polygons = polygons
        self.source = source
        self.receiver = receiver

    def plot(self, ax: matplotlib.axes.Axes = None):
        """Plot the environment."""
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        i = 0
        self.source.plot(ax)
        self.receiver.plot(ax)
        for i in range(len(self.polygons)):
            self.polygons[i].plot(ax, colors[i])
        ax.set_xlabel('x',labelpad=30)
        ax.set_ylabel('y',labelpad=30)
        ax.set_zlabel('z')
        ax.set_aspect('equal', 'box')


def _cmp_floats(a, b, atol=1e-12):
    return abs(a-b) < atol


def _magnitude(vector):
    return np.sqrt(np.dot(np.array(vector), np.array(vector)))


def _norm(vector):
    return np.array(vector)/_magnitude(np.array(vector))


####################################################
# patch processing (numpy formulation)
def _process_patches(
        polygon_points_array: np.ndarray,
        walls_normal: np.ndarray,
        patch_size:float, n_walls:int):
    """Process the patches.

    Parameters
    ----------
    polygon_points_array : np.ndarray
        points of the polygon of shape (n_walls, 4, 3)
    walls_normal : np.ndarray
        wall normal of shape (n_walls, 3)
    patch_size : float
        maximal patch size in meters of shape (n_walls, 3).
    n_walls : int
        number of walls

    Returns
    -------
    patches_points : np.ndarray
        points of all patches of shape (n_patches, 4, 3)
    patches_normal : np.ndarray
        normal of all patches of shape (n_patches, 3)
    n_patches : int
        number of patches

    """
    n_patches = 0
    n_walls = polygon_points_array.shape[0]
    for i in range(n_walls):
        n_patches += _total_number_of_patches(
            polygon_points_array[i, :, :], patch_size)
    patches_points = np.empty((n_patches, 4, 3))
    patch_to_wall_ids = np.empty((n_patches), dtype=np.int64)
    patches_per_wall = np.empty((n_walls), dtype=np.int64)

    for i in range(n_walls):
        polygon_points = polygon_points_array[i, :]
        patches_points_wall = _create_patches(
            polygon_points, patch_size)
        patches_per_wall[i] = patches_points_wall.shape[0]
        j_start = (np.sum(patches_per_wall[:i])) if i > 0 else 0
        j_end = np.sum(patches_per_wall[:i+1])
        patch_to_wall_ids[j_start:j_end] = i
        patches_points[j_start:j_end, :, :] = patches_points_wall
    n_patches = patches_points.shape[0]

    # calculate patch information
    patches_normal = walls_normal[patch_to_wall_ids, :]
    return (patches_points, patches_normal, n_patches, patch_to_wall_ids)

def _total_number_of_patches(polygon_points:np.ndarray, max_size: float):
    """Calculate the total number of patches.

    Parameters
    ----------
    polygon_points : np.ndarray
        points of the polygon of shape (4, 3)
    max_size : float
        maximal patch size in meters

    Returns
    -------
    n_patches : int
        number of patches

    """
    size = np.empty(polygon_points.shape[1])
    for i in range(polygon_points.shape[1]):
        size[i] = polygon_points[:, i].max() - polygon_points[:, i].min()
    patch_nums = np.array([int(n) for n in size/max_size])
    if patch_nums[2] == 0:
        x_idx = 0
        y_idx = 1
    if patch_nums[1] == 0:
        x_idx = 0
        y_idx = 2
    if patch_nums[0] == 0:
        x_idx = 1
        y_idx = 2

    return patch_nums[x_idx]*patch_nums[y_idx]

def _create_patches(polygon_points:np.ndarray, max_size):
    """Create patches from a polygon."""
    size = np.empty(polygon_points.shape[1])
    for i in range(polygon_points.shape[1]):
        size[i] = polygon_points[:, i].max() - polygon_points[:, i].min()
    patch_nums = np.array([int(n) for n in size/max_size])
    real_size = size/patch_nums
    if patch_nums[2] == 0:
        x_idx = 0
        y_idx = 1
    if patch_nums[1] == 0:
        x_idx = 0
        y_idx = 2
    if patch_nums[0] == 0:
        x_idx = 1
        y_idx = 2

    x_min = np.min(polygon_points.T[x_idx])
    y_min = np.min(polygon_points.T[y_idx])

    n_patches = patch_nums[x_idx]*patch_nums[y_idx]
    patches_points = np.empty((n_patches, 4, 3))
    i = 0
    for i_x in range(patch_nums[x_idx]):
        for i_y in range(patch_nums[y_idx]):
            points = polygon_points.copy()
            points[0, x_idx] = x_min + i_x * real_size[x_idx]
            points[0, y_idx] = y_min + i_y * real_size[y_idx]
            points[1, x_idx] = x_min + (i_x+1) * real_size[x_idx]
            points[1, y_idx] = y_min + i_y * real_size[y_idx]
            points[3, x_idx] = x_min + i_x * real_size[x_idx]
            points[3, y_idx] = y_min + (i_y+1) * real_size[y_idx]
            points[2, x_idx] = x_min + (i_x+1) * real_size[x_idx]
            points[2, y_idx] = y_min + (i_y+1) * real_size[y_idx]
            patches_points[i] = points
            i += 1

    return patches_points

def _calculate_center(points):
    return np.sum(points, axis=-2) / points.shape[-2]

def _calculate_size(points):
    vec1 = points[..., 0, :]-points[..., 1, :]
    vec2 = points[..., 1, :]-points[..., 2, :]
    return np.abs(vec1-vec2)

def _polygon_area(pts: np.ndarray) -> float:
    """Calculate the area of a convex n-sided polygon.

    Parameters
    ----------
    pts: np.ndarray
        list of 3D points which define the vertices of the polygon

    Returns
    -------
    area: float
        area of polygon defined by pts

    """
    area = 0

    for tri in range(pts.shape[0]-2):
        area  +=  .5 * np.linalg.norm(
                            np.cross(pts[tri+1] - pts[0], pts[tri+2]-pts[0]) )

    return area

def _calculate_area(points):
    area = np.zeros(points.shape[0])

    for i in prange(points.shape[0]):
        area[i] = _polygon_area(points[i])

    return area

####################################################
# point, vector, patch operations
def _calculate_normals(points: np.ndarray):
    """Calculate normals of plane defined by 3 or more input points.

    Parameters
    ----------
    points: np.ndarray (n_planes, n_points, 3)
        collection of points on common planes.

    Returns
    -------
    normals: np.ndarray (n_planes,3)
        collection of normal vectors to planes defined by point collection.
    """

    normals = np.empty((points.shape[0],3))

    for i in prange(points.shape[0]):
        normals[i]=np.cross(points[i][1]-points[i][0],points[i][2]-points[i][0])
        normals[i]/=np.linalg.norm(normals[i])

    return normals

def _matrix_vector_product(matrix: np.ndarray,vector:np.ndarray)->np.ndarray:
    """Compute the inner product between a matrix and a vector to please njit.

    Parameters
    ----------
    matrix : numpy.ndarray(n,n)
        input matrix

    vector : numpy.ndarray(n,)
        vector

    Returns
    -------
    out: np.ndarray(n,)
        matrix*vector inner product

    """
    out = np.empty(matrix.shape[0])

    for i in prange(matrix.shape[0]):
        out[i] = np.dot(matrix[i],vector)

    return out

def _rotation_matrix(n_in: np.ndarray, n_out=np.array([])):
    """Compute a rotation matrix from a given input and output directions.

    Parameters
    ----------
    n_in : numpy.ndarray(3,)
        input vector

    n_out : numpy.ndarray(3,)
        direction to which n_in is to be rotated

    Returns
    -------
    matrix: np.ndarray
        rotation matrix

    """
    if n_out.shape[0] == 0:
        n_out = np.zeros_like(n_in)
        n_out[-1] = 1.
    else:
        n_out = n_out

    #check if all the vector entries coincide
    counter = int(0)

    for i in prange(n_in.shape[0]):
        if n_in[i] == n_out[i]:
            counter+=1
        else:
            counter=counter

    # if input vector is the same as output return identity matrix
    if counter == n_in.shape[0]:

        matrix = np.eye( len(n_in) , dtype=np.float64)

    else:

        a = n_in / np.linalg.norm(n_in)
        a = np.reshape(a, len(n_in) )

        b = n_out / np.linalg.norm(n_out)
        b = np.reshape(b, len(n_in) )

        c = np.dot(a,b)

        if c!=-1:
            v = np.cross(a,b)
            s = np.linalg.norm(v)
            kmat = np.array([[0, -v[2], v[1]],
                             [v[2], 0, -v[0]],
                             [-v[1], v[0], 0]])

            matrix =  ( np.eye( len(n_in) ) +
                       kmat +
                       kmat.dot(kmat) * ((1 - c) / (s ** 2)) )

        else: # in case the in and out vectors have symmetrical directions
            matrix = np.array([[-1.,0.,0.],[0.,1.,0.],[0.,0.,-1.]])

    return matrix

def _project_to_plane(origin: np.ndarray, point: np.ndarray,
                     plane_pt: np.ndarray, plane_normal: np.ndarray,
                     epsilon=1e-6, check_normal=True):
    """Project point onto plane following direction defined by origin.

    Also applicable to lower or higher-dimensional spaces, not just 3D.

    Parameters
    ----------
    origin: np.ndarray(float) (n_dims,)
        point determining projection direction

    point: np.ndarray(float) (n_dims,)
        point to project

    plane_pt: np.ndarray(float) (n_dims,)
        reference point on projection plane

    plane_normal: np.ndarray(float) (n_dims,)
        normal of projection plane

    epsilon: float
        tolerance for plane orthogonality check

    check_normal: bool
        if True, discards projections on planes facing
        opposite direction from the point and origin.

    Returns
    -------
    int_point: np.ndarray(float) (n_dims,) or None
        intersection point.
        None if no intersection point is found

    """
    v = point-origin
    dotprod = np.dot(v,plane_normal)

    cond = dotprod < -epsilon

    if not check_normal:
        cond = np.abs(dotprod) > epsilon

    if cond:
        w = point-plane_pt
        fac = -np.divide(np.dot(plane_normal,w), dotprod)

        int_point = w + plane_pt + fac*v
    else:
        int_point = None

    return int_point

def _point_in_polygon(point3d: np.ndarray,
                     polygon3d: np.ndarray, plane_normal: np.ndarray,
                     eta=1e-6) -> bool:
    """Check if point is inside given polygon.

    Parameters
    ----------
    point3d: np.ndarray(float) (3,)
        point being evaluated.

    polygon3d: np.ndarray (N,3)
        polygon boundary points

    plane_normal: np.ndarray(float) (3,)
        normal of the polygon's plane

    eta: float
        tolerance for point inside of line check.

    Returns
    -------
    out: bool
        flags if point is inside polygon
        (True if inside, False if not)

    """
    # check coplanarity of polygon and point3D
    if np.abs(np.dot(point3d-polygon3d[0],plane_normal))>eta:
        out= False
    else:

        # rotate all (coplanar) points to a horizontal plane
        # and remove z dimension for convenience
        rotmat = _rotation_matrix(n_in=plane_normal)

        pt = _matrix_vector_product(matrix=rotmat,
                                    vector=point3d)[0:point3d.shape[0]-1]

        poly = np.empty((polygon3d.shape[0],2))
        for i in prange(polygon3d.shape[0]):
            poly[i] = _matrix_vector_product(
                matrix=rotmat,vector=polygon3d[i])[0:point3d.shape[0]-1]


        # winding number algorithm
        count = 0
        for i in prange(poly.shape[0]):
            a1 = poly[(i+1)%poly.shape[0]]
            a0 = poly[i%poly.shape[0]]
            side = a1-a0

            # check if line from evaluation point in +x direction
            # intersects polygon side
            nl = np.array([-side[1],side[0]])/np.linalg.norm(side)
            b = _project_to_plane(origin=pt, point=pt+np.array([1.,0.]),
                                plane_pt=a1, plane_normal=nl,
                                check_normal=False)

            # check if intersection exists and if is inside side [a0,a1]
            if (b is not None) and b[0]>pt[0]:
                if abs(np.linalg.norm(b-a0)+np.linalg.norm(b-a1)
                                    -np.linalg.norm(a1-a0)) <= eta:
                    if np.dot(b-pt,nl)>0:
                        count+=1
                    elif np.dot(b-pt,nl)<0:
                        count-=1

        if count != 0:
            out = True
        else:
            out = False

    return out

def _sphere_tangent_vector(v0: np.ndarray, v1:np.ndarray) -> np.ndarray:
    """Compute a vector tangent to a spherical surface based on two points.

    The tangent vector is evaluated on point v0.
    The respective tangent arc on the sphere joins points v0 and v1.

    Parameters
    ----------
    v0 : numpy.ndarray(3,)
        point on which to calculate tangent vector

    v1 : numpy.ndarray(3,)
        point on sphere to which tangent vector "points"

    Returns
    -------
    vout: np.ndarray
        vector tangent to spherical surface

    """
    if np.abs(np.dot(v0,v1))>1e-10:
        vout = (v1-v0)-np.dot((v1-v0),v0)/np.dot(v0,v0)*v0
        vout /= np.linalg.norm(vout)

    else:
        vout = v1/np.linalg.norm(v1)

    return vout

####################################################
# checks
def _coincidence_check(p0: np.ndarray, p1: np.ndarray, thres = 1e-6) -> bool:
    """Flag true if two lists of points have any common points.

    Parameters
    ----------
    p0 : numpy.ndarray(# vertices, 3)
        patch

    p1 : numpy.ndarray(# vertices, 3)
        another patch

    thres: float
        threshold value for distance between points

    Returns
    -------
    flag: bool
        flag of coincident points on both patches

    """
    flag = False

    for i in prange(p0.shape[0]):
        for j in prange(p1.shape[0]):
            if np.linalg.norm(p0[i]-p1[j])<thres:
                flag=True
                break


    return flag

def _check_patch2patch_visibility(
        patches_center:np.ndarray,
        surf_normal:np.ndarray, surf_points:np.ndarray) -> np.ndarray:
    """Check the visibility between patches.

    Parameters
    ----------
    patches_center : np.ndarray
        center points of all patches of shape (n_patches, 3)
    surf_normal : np.ndarray
        normal vectors of all patches of shape (n_patches, 3)
    surf_points : np.ndarray
        boundary points of possible blocking surfaces  (n_surfaces,n_patches,3)

    Returns
    -------
    visibility_matrix : np.ndarray
        boolean matrix of shape (n_patches, n_patches) with True if patches
        can see each other, otherwise false

    """
    n_patches = patches_center.shape[0]
    visibility_matrix = np.empty((n_patches, n_patches), dtype=np.bool_)
    visibility_matrix.fill(False)
    indexes = []
    for i_source in range(n_patches):
        for i_receiver in range(n_patches):
            if i_source < i_receiver:
                indexes.append((i_source, i_receiver))
                visibility_matrix[i_source,i_receiver]=True
    indexes = np.array(indexes)
    for i in prange(indexes.shape[0]):
        i_source = indexes[i, 0]
        i_receiver = indexes[i, 1]

        surfid=0
        while (visibility_matrix[i_source, i_receiver] and
                                surfid!=len(surf_normal)):

            visibility_matrix[i_source, i_receiver]= _basic_visibility(
                                                        patches_center[i_source],
                                                        patches_center[i_receiver],
                                                        surf_points[surfid],
                                                        surf_normal[surfid])

            surfid+=1

    return visibility_matrix

def _check_point2patch_visibility(
        eval_point:np.ndarray,
        patches_center:np.ndarray,
        surf_normal:np.ndarray, surf_points:np.ndarray) -> np.ndarray:
    """Check the visibility between a point and patches.

    It is intended to check the visibiliy between source/receivers and patches.

    Parameters
    ----------
    eval_point : np.ndarray
        evaluation point for visibility of shape (3)
    patches_center : np.ndarray
        center points of all patches of shape (n_patches, 3)
    surf_normal : np.ndarray
        normal vectors of all patches of shape (n_patches, 3)
    surf_points : np.ndarray
        boundary points of possible blocking surfaces (n_surfaces,n_patches,3)

    Returns
    -------
    visibility_vector : np.ndarray
        boolean vector of shape (n_patches) with True if patches
        are visible from the evaluation point, otherwise false

    """
    n_patches = patches_center.shape[0]
    visibility_vector = np.ones((n_patches), dtype=np.bool_)
    for i in prange(n_patches):
        surfid=0
        while (visibility_vector[i] and
                                surfid!=len(surf_normal)):

            visibility_vector[i]= _basic_visibility(eval_point,
                                                    patches_center[i],
                                                    surf_points[surfid],
                                                    surf_normal[surfid])

            surfid+=1

    return visibility_vector

def _basic_visibility(vis_point: np.ndarray,
                     eval_point: np.ndarray,
                     surf_points: np.ndarray,
                     surf_normal: np.ndarray,
                     eta=1e-6)->bool:
    """Return visibility of a point based on view point position.

    Parameters
    ----------
    vis_point: np.ndarray (3,)
        view point from which the visibility is being evaluated.

    eval_point: np.ndarray (3,)
        point being evaluated for visibility.

    surf_points: np.ndarray (N,3)
        boundary points of a possibly blocking surface.

    surf_normal: np.ndarray (3,)
        normal of possibly blocking surface.

    eta: float
        coplanarity check tolerance

    Returns
    -------
    is_visible: bool
        point visibility flag.
        (1 if points are visible to eachother, otherwise 0)

    """
    is_visible = True

    vis_in_surf = _point_in_polygon(point3d=vis_point,
                                    polygon3d=surf_points,
                                    plane_normal=surf_normal)
    eval_in_surf = _point_in_polygon(point3d=eval_point,
                                    polygon3d=surf_points,
                                    plane_normal=surf_normal)

    if not vis_in_surf and not eval_in_surf:

        pt = _project_to_plane(origin=vis_point, point=eval_point,
                            plane_pt=surf_points[0],
                            plane_normal=surf_normal,
                            check_normal=False)

        if pt is not None:
            if _point_in_polygon(point3d=pt,
                                 polygon3d=surf_points,
                                 plane_normal=surf_normal):
                if (np.dot(pt-vis_point,pt-eval_point)<0):
                    is_visible = False

    elif (vis_in_surf and not eval_in_surf and
          np.dot(surf_normal, eval_point-vis_point)<0):
        is_visible=False

    elif (not vis_in_surf and eval_in_surf and
          np.dot(surf_normal, vis_point-eval_point)<0):
            is_visible=False

    elif (np.abs(np.dot(vis_point-surf_points[0],surf_normal))<eta and
          np.abs(np.dot(eval_point-surf_points[0],surf_normal))<eta and
          (vis_in_surf or eval_in_surf)):
        is_visible=False


    return is_visible


if numba is not None:
    _total_number_of_patches = numba.njit()(_total_number_of_patches)
    _process_patches = numba.njit()(_process_patches)
    _calculate_area = numba.njit()(_calculate_area)
    _calculate_center = numba.njit()(_calculate_center)
    _calculate_size = numba.njit()(_calculate_size)
    _create_patches = numba.njit()(_create_patches)
    _check_patch2patch_visibility = numba.njit(parallel=True)(
        _check_patch2patch_visibility)
    _check_point2patch_visibility = numba.njit(parallel=True)(
        _check_point2patch_visibility)
    _calculate_normals = numba.njit()(_calculate_normals)
    _coincidence_check = numba.njit()(_coincidence_check)
    _basic_visibility = numba.njit()(_basic_visibility)
    _project_to_plane = numba.njit()(_project_to_plane)
    _point_in_polygon = numba.njit()(_point_in_polygon)
    _matrix_vector_product = numba.njit()(_matrix_vector_product)
    _rotation_matrix = numba.njit()(_rotation_matrix)
    _polygon_area = numba.njit()(_polygon_area)
    _sphere_tangent_vector = numba.njit()(_sphere_tangent_vector)





"""Contains tests for the geometry module.

Geometry and wall directivity functionality in the radiosity package.
"""
import numpy as np
import numpy.testing as npt
import sparrowpy.geometry as geo
from sparrowpy.sound_object import Receiver, SoundSource


def test_polygon_defaults():
    """Test Polygon class."""
    poly = geo.Polygon(
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], [0, 1, 0], [0, 0, 1])
    npt.assert_equal(poly.pts[0], np.array([0, 0, 0]))
    npt.assert_equal(poly.pts[1], np.array([1, 0, 0]))
    npt.assert_equal(poly.pts[2], np.array([1, 1, 0]))
    npt.assert_equal(poly.pts[3], np.array([0, 1, 0]))
    npt.assert_almost_equal(poly.up_vector, np.array([0, 1, 0]))
    npt.assert_allclose(poly.normal, np.array([0, 0, 1]))


def test_polygon_center():
    """Test center property of Polygon."""
    poly = geo.Polygon(
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], [0, 1, 0], [0, 0, 1])
    npt.assert_allclose(poly.center, np.array([0.5, .5, 0]))

def test_polygon_size():
    """Test size property of Polygon."""
    poly = geo.Polygon(
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], [0, 1, 0], [0, 0, 1])
    npt.assert_allclose(poly.size, np.array([1, 1, 0]))


def test_polygon_surface_normal():
    """Test surface_normal property of Polygon."""
    poly = geo.Polygon(
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], [0, 1, 0], [0, 0, 1])
    npt.assert_equal(poly.normal, np.array([0, 0, 1]))
    npt.assert_equal(poly.up_vector, np.array([0, 1, 0]))


def test_polygon_n_points():
    """Test n_points property of Polygon."""
    poly = geo.Polygon(
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], [0, 1, 0], [0, 0, 1])
    assert poly.n_points == 4


def test_polygon_intersection():
    """Test intersection method of Polygon."""
    poly = geo.Polygon(
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], [0, 1, 0], [0, 0, 1])
    inter = poly.intersection([0.5, 0.5, 1], [0, 0, -1])
    npt.assert_allclose(inter, np.array([0.5, 0.5, 0]))
    npt.assert_allclose(poly.normal, np.array([0, 0, 1]))


def test_polygon_no_intersection():
    """Test intersection method of Polygon."""
    poly = geo.Polygon(
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], [0, 1, 0], [0, 0, 1])
    inter = poly.intersection([0.5, 0.5, 1], [0, 0, 1])
    assert inter is None
    npt.assert_allclose(poly.normal, np.array([0, 0, 1]))


def test_environment():
    """Test Environment class."""
    speed_of_sound = 346.2
    polygon = geo.Polygon(
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], [0, 1, 0], [0, 0, 1])
    source = SoundSource([0, 0, 0], [1, 0, 0], [0, 0, 1])
    receiver = Receiver([0, 0, 0], [1, 0, 0], [0, 0, 1])
    # test all parameter
    env = geo._Environment([polygon], source, receiver, speed_of_sound)
    assert env.polygons == [polygon]
    assert env.speed_of_sound == speed_of_sound
    assert env.receiver == receiver
    assert env.source == source


def test_environment_defaults():
    """Test Environment class."""
    polygon = geo.Polygon(
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], [0, 1, 0], [0, 0, 1])
    source = SoundSource([0, 0, 0], [1, 0, 0], [0, 0, 1])
    receiver = Receiver([0, 0, 0], [1, 0, 0], [0, 0, 1])
    # test all parameter
    env = geo._Environment([polygon], source, receiver, speed_of_sound=346.18)
    assert env.polygons == [polygon]
    assert env.speed_of_sound == 346.18
    assert env.receiver == receiver
    assert env.source == source


def test_polygon_to_dict():
    """Test to_dict method of Polygon."""
    # Arrange
    points = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]])
    up_vector = np.array([0, 1, 0])
    normal = np.array([0, 0, 1])
    polygon = geo.Polygon(points, up_vector, normal)

    # Act
    result = polygon.to_dict()

    # Assert
    assert isinstance(result, dict)
    assert 'pts' in result
    assert 'up_vector' in result
    assert 'normal' in result
    np.testing.assert_array_equal(result['pts'], points)
    np.testing.assert_array_equal(result['up_vector'], up_vector)
    np.testing.assert_array_equal(result['normal'], normal)


def test_polygon_from_dict():
    """Test from_dict method of Polygon."""
    # Arrange
    points = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]])
    up_vector = np.array([0, 1, 0])
    normal = np.array([0, 0, 1])
    polygon_dict = {
        'pts': points.tolist(),
        'up_vector': up_vector.tolist(),
        'normal': normal.tolist(),
    }

    # Act
    polygon = geo.Polygon.from_dict(polygon_dict)

    # Assert
    assert isinstance(polygon, geo.Polygon)
    np.testing.assert_array_equal(polygon.pts, points)
    np.testing.assert_array_equal(polygon.up_vector, up_vector)
    np.testing.assert_array_equal(polygon._normal, normal)



"""Test sparrowpy package."""
import pytest


def test_import_sparrowpy():
    """Test importing sparrowpy."""
    try:
        import sparrowpy           # noqa
    except ImportError:
        pytest.fail('import sparrowpy failed')



"""SoundObject class for spatial audio reproduction."""
import matplotlib
import numpy as np
import pyfar as pf
import sofar as sf


class DirectivityMS():
    """Directivity class for FreeFieldDirectivityTF convention."""

    data: pf.FrequencyData
    receivers: pf.Coordinates

    def __init__(self, file_path: str, source_index=0) -> None:
        """Init DirectivityMS.

        Parameters
        ----------
        file_path : str
            directivity path for sofa file.
        source_index : int, optional
            source index of directivity, by default 0

        """
        sofa = sf.read_sofa(file_path, verbose=False)
        self.data = pf.FrequencyData(
            sofa.Data_Real[source_index, :] + 1j * sofa.Data_Imag[
                source_index, :], sofa.N)
        positions = sofa.ReceiverPosition
        if positions.ndim > 2:
            positions = np.squeeze(positions)
        self.receivers = pf.io.io._sofa_pos(
            sofa.ReceiverPosition_Type, positions)
        if  self.receivers.cdim != 1:
            raise ValueError(
                'DirectivityMS only supports 1D coordinates, '
                f'got {self.receivers.cdim}D coordinates. Squeezing did not '
                'work.')

    def get_directivity(
            self, source_pos: np.ndarray, source_view: np.ndarray,
            source_up: np.ndarray, target_position: np.ndarray,
            i_freq: int) -> float:
        """Get Directivity for certain position.

        Parameters
        ----------
        source_pos : np.ndarray
            cartesian source position in m
        source_view : np.ndarray
            cartesian source view in m
        source_up : np.ndarray
            cartesian source up in m
        target_position : np.ndarray
            cartesian target position in m
        i_freq : int
            frequency bin index

        Returns
        -------
        float
            nearest directivity factor for given position and orientation.

        """
        (azimuth_deg, elevation_deg) = _get_metrics(
            source_pos, source_view, source_up, target_position)
        find = pf.Coordinates.from_spherical_elevation(
            azimuth_deg/180*np.pi, elevation_deg/180*np.pi, 1,
        )
        index, _ = self.receivers.find_nearest(find)
        return self.data.freq[index, i_freq]


def _get_metrics(pos_G, view_G, up_G, target_pos_G):
    pos_G = np.array(pos_G, dtype=float)
    view_G = np.array(view_G, dtype=float)
    up_G = np.array(up_G, dtype=float)
    target_pos_G = np.array(target_pos_G, dtype=float)
    direction_G = target_pos_G - pos_G

    x_dash = np.cross(view_G, up_G)
    y_dash = up_G
    z_dash = -view_G

    w_x_local = np.dot(direction_G, x_dash)
    w_y_local = np.dot(direction_G, y_dash)
    w_z_local = np.dot(direction_G, z_dash)

    azimuth_deg = np.arctan2(-w_x_local, -w_z_local) / np.pi * 180
    w = np.array([w_x_local, w_y_local, w_z_local])
    w_y_local_normalized = w_y_local / np.sqrt(np.dot(w, w))
    elevation_deg = np.arcsin(w_y_local_normalized) / np.pi * 180

    return (azimuth_deg, elevation_deg)


class SoundObject():
    """A class holding the common properties for Source and Receiver."""

    position: np.ndarray
    view: np.ndarray
    up: np.ndarray

    def __init__(
            self, position: np.ndarray, view: np.ndarray,
            up: np.ndarray) -> None:
        """Init a sound object.

        Parameters
        ----------
        position : np.ndarray
            position of the sound object in m
        view : np.ndarray
            view vector of sound object
        up : np.ndarray
            uo vector of sound object

        """
        self.position = np.array(position, dtype=float)
        assert self.position.shape == (3,)
        self.view = np.array(view, dtype=float)
        self.view /= np.sqrt(np.dot(view, view))
        assert self.view.shape == (3,)
        self.up = np.array(up, dtype=float)
        self.up /= np.sqrt(np.dot(up, up))
        assert self.up.shape == (3,)

    def plot(self, ax: matplotlib.axes.Axes, **kwargs):
        """Plot SoundObject position and orientation.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes to plot on.
        **kwargs
            Keyword arguments that are passed to
            ``matplotlib.pyplot.scatter()``.

        """
        xyz = self.position
        ax.scatter(xyz[0], xyz[1], xyz[2], kwargs)





class SoundSource(SoundObject):
    """Acoustic sound source inhered from SoundObject."""

    directivity: DirectivityMS
    sound_power: float

    def __init__(
            self, position: np.ndarray, view: np.ndarray,
            up: np.ndarray, directivity: DirectivityMS = None,
            sound_power: float = 1) -> None:
        """Init sound source.

        Parameters
        ----------
        position : np.ndarray
            position of the sound source in m
        view : np.ndarray
            view vector of sound source
        up : np.ndarray
            uo vector of sound source
        directivity : DirectivityMS, optional
            Directivity, by default None
        sound_power : float, optional
            sound power of the source in Watt, by default 1

        """
        super(SoundSource, self).__init__(position, view, up)
        self.sound_power = float(sound_power)
        if directivity is not None:
            assert isinstance(directivity, DirectivityMS)
        self.directivity = directivity

    def plot(self, ax, **kwargs):
        """Plot Source position and orientation.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes to plot on.
        **kwargs
            Keyword arguments that are passed to
            ``matplotlib.pyplot.scatter()``.

        """
        super(SoundSource, self).plot(ax, color='r', label='Source', **kwargs)

    def get_directivity(
            self, target_position: np.ndarray, frequency: float) -> float:
        """Get Directivity for certain position and frequency.

        Parameters
        ----------
        target_position : np.ndarray
            cartesian target position in m of shape (3,) or (n, 3).
        frequency : float
            frequency in Hz.

        Returns
        -------
        float
            nearest directivity factor for given position and frequency.
        """
        i_freq = np.argmin(np.abs(self.directivity.data.frequencies-frequency))
        if target_position.size == 3:
            return self.directivity.get_directivity(
                self.position, self.view, self.up, target_position, i_freq)
        else:
            return np.array([
                self.directivity.get_directivity(
                    self.position, self.view, self.up, pos, i_freq)
                for pos in target_position
            ])[:, 0]

class Receiver(SoundObject):
    """Receiver object inhered from SoundObject."""

    def __init__(
            self, position: np.ndarray, view: np.ndarray,
            up: np.ndarray) -> None:
        """Init sound receiver.

        Parameters
        ----------
        position : np.ndarray
            cartesian positions for receiver.
        view : np.ndarray
            view vector of sound receiver.
        up : np.ndarray
            up vector of sound receiver.

        """
        super(Receiver, self).__init__(position, view, up)

    def plot(self, ax, **kwargs):
        """Plot Receiver position and orientation.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes to plot on.
        **kwargs
            Keyword arguments that are passed to
            ``matplotlib.pyplot.scatter()``.

        """
        super(Receiver, self).plot(ax, color='b', label='Receiver', **kwargs)



"""Form factor calculation for radiosity."""
try:
    import numba
    prange = numba.prange
except ImportError:
    numba = None
    prange = range
import numpy as np

## form factor matrix assembly

def patch2patch_ff_kang(
        patches_center:np.ndarray, patches_normal:np.ndarray,
        patches_size:np.ndarray, visible_patches:np.ndarray) -> np.ndarray:
    """Calculate the form factors between patches.

    Parameters
    ----------
    patches_center : np.ndarray
        center points of all patches of shape (n_patches, 3)
    patches_normal : np.ndarray
        normal vectors of all patches of shape (n_patches, 3)
    patches_size : np.ndarray
        size of all patches of shape (n_patches, 3)
    visible_patches : np.ndarray
        index list of all visible patches combinations (n_combinations, 2)

    Returns
    -------
    form_factors : np.ndarray
        form factors between all patches of shape (n_patches, n_patches)
        note that just i_source < i_receiver are calculated ff[i, j] = ff[j, i]

    """
    n_patches = patches_center.shape[0]
    form_factors = np.zeros((n_patches, n_patches))
    for i in prange(visible_patches.shape[0]):
        i_source = int(visible_patches[i, 0])
        i_receiver = int(visible_patches[i, 1])
        source_center = patches_center[i_source]
        source_normal = patches_normal[i_source]
        receiver_center = patches_center[i_receiver]
        # calculation of form factors
        receiver_normal = patches_normal[i_receiver]
        dot_product = np.dot(receiver_normal, source_normal)

        if dot_product == 0:  # orthogonal

            if np.abs(source_normal[0]) > 1e-5:
                idx_source = {2, 1}
                dl = source_center[2]
                dm = source_center[1]
                dd_l = patches_size[i_source, 2]
                dd_m = patches_size[i_source, 1]
            elif np.abs(source_normal[1]) > 1e-5:
                idx_source = {2, 0}
                dl = source_center[2]
                dm = source_center[0]
                dd_l = patches_size[i_source, 2]
                dd_m = patches_size[i_source, 0]
            elif np.abs(source_normal[2]) > 1e-5:
                idx_source = {0, 1}
                dl = source_center[1]
                dm = source_center[0]
                dd_l = patches_size[i_source, 1]
                dd_m = patches_size[i_source, 0]

            if np.abs(receiver_normal[0]) > 1e-5:
                idx_l = 1 if 1 in idx_source else 2
                idx_s = 0
                idx_r = 2 if 1 in idx_source else 1
                dl_prime = receiver_center[1]
                dn_prime = receiver_center[2]
            elif np.abs(receiver_normal[1]) > 1e-5:
                idx_l = 0 if 0 in idx_source else 2
                idx_s = 1
                idx_r = 2 if 0 in idx_source else 0
                dl_prime = receiver_center[0]
                dn_prime = receiver_center[2]
            elif np.abs(receiver_normal[2]) > 1e-5:
                idx_l = 0 if 0 in idx_source else 1
                idx_s = 2
                idx_r = 1 if 0 in idx_source else 0
                dl_prime = receiver_center[1]
                dn_prime = receiver_center[0]

            dm = np.abs(
                source_center[idx_s]-receiver_center[idx_s])
            dl = source_center[idx_l]
            dl_prime = receiver_center[idx_l]
            dn_prime = np.abs(
                source_center[idx_r]-receiver_center[idx_r])

            d = np.sqrt( ( (dl - dl_prime) ** 2 ) + ( dm ** 2 ) + (
                dn_prime ** 2) )

            # Equation 13
            A = ( dm - ( 0.5 * dd_m ) ) / ( np.sqrt( ( (
                dl - dl_prime) ** 2 ) + ( ( dm - (
                    0.5 * dd_m ) ) ** 2 ) + ( dn_prime ** 2) ) )

            # Equation 14
            B_num = dm + (0.5 * dd_m)
            B_denum =  np.sqrt(( np.square(dl - dl_prime) ) + (
                np.square(dm + (0.5*dd_m)) ) + (
                    np.square(dn_prime)) )
            B = B_num/B_denum

            # Equation 15
            one = np.arctan( np.abs( ( dl - (
                0.5*dd_l) - dl_prime ) / (dn_prime) ) )
            two = np.arctan( np.abs( ( dl + (
                0.5*dd_l) - dl_prime ) / (dn_prime) ) )

            k = -1 if np.abs(dl - dl_prime) < 1e-12 else 1

            theta = np.abs( one - (k*two) )

            # Equation 11
            ff =  (
                1 / (2 * np.pi) ) * (np.abs(
                    (A ** 2) - (B ** 2) )) * theta

        else:
            # parallel
            if np.abs(receiver_normal[0]) > 1e-5:
                dl = receiver_center[1]
                dm = receiver_center[0]
                dn = receiver_center[2]
                dl_prime = source_center[1]
                dm_prime = source_center[0]
                dn_prime = source_center[2]
                dd_l = patches_size[i_source, 1]
                if patches_size.shape[1] > 2:
                    dd_n = patches_size[i_source, 2]
                else:
                    dd_n = patches_size[i_source, 1]
            elif np.abs(receiver_normal[1]) > 1e-5:
                dl = receiver_center[0]
                dm = receiver_center[1]
                dn = receiver_center[2]
                dl_prime = source_center[0]
                dm_prime = source_center[1]
                dn_prime = source_center[2]
                dd_l = patches_size[i_source, 0]
                if patches_size.shape[1] > 2:
                    dd_n = patches_size[i_source, 2]
                else:
                    dd_n = patches_size[i_source, 1]
            elif np.abs(receiver_normal[2]) > 1e-5:
                dl = receiver_center[1]
                dm = receiver_center[2]
                dn = receiver_center[0]
                dl_prime = source_center[1]
                dm_prime = source_center[2]
                dn_prime = source_center[0]
                dd_l = patches_size[i_source, 1]
                dd_n = patches_size[i_source, 0]

            d = np.sqrt(
                ( (dl - dl_prime) ** 2 ) +
                ( (dn - dn_prime) ** 2 ) +
                ( (dm - dm_prime) ** 2 ) )
            # Equation 16
            ff =  ( dd_l * dd_n * ( (
                dm-dm_prime) ** 2 ) ) / ( np.pi * ( d**4 ) )

        form_factors[i_source, i_receiver] = ff
    return form_factors

def _source2patch_energy_kang(
        source_position: np.ndarray, patches_center: np.ndarray,
        patches_normal: np.ndarray, air_attenuation:np.ndarray,
        patches_size: float, n_bins:float):
    """Calculate the initial energy from the source.

    Parameters
    ----------
    source_position : np.ndarray
        source position of shape (3,)
    patches_center : np.ndarray
        center of all patches of shape (n_patches, 3)
    patches_normal : np.ndarray
        normal of all patches of shape (n_patches, 3)
    air_attenuation : np.ndarray
        air attenuation factor in Np/m (n_bins,)
    patches_size : float
        size of all patches of shape (n_patches, 3)
    n_bins : float
        number of frequency bins.

    Returns
    -------
    energy : np.ndarray
        energy of all patches of shape (n_patches)
    distance : np.ndarray
        corresponding distance of all patches of shape (n_patches)

    """
    n_patches = patches_center.shape[0]
    energy = np.empty((n_patches, n_bins))
    distance_out = np.empty((n_patches, ))
    for j in prange(n_patches):
        source_pos = source_position.copy()
        receiver_pos = patches_center[j, :].copy()
        receiver_normal = patches_normal[j, :].copy()
        receiver_size = patches_size[j, :].copy()

        if np.abs(receiver_normal[2]) > 0.99:
            i = 2
            indexes = [0, 1, 2]
        elif np.abs(receiver_normal[1]) > 0.99:
            indexes = [2, 0, 1]
            i = 1
        elif np.abs(receiver_normal[0]) > 0.99:
            i = 0
            indexes = [1, 2, 0]
        offset = receiver_pos[i]
        source_pos[i] = np.abs(source_pos[i] - offset)
        receiver_pos[i] = np.abs(receiver_pos[i] - offset)
        dl = receiver_pos[indexes[0]]
        dm = receiver_pos[indexes[1]]
        dn = receiver_pos[indexes[2]]
        dd_l = receiver_size[indexes[0]]
        dd_m = receiver_size[indexes[1]]
        S_x = source_pos[indexes[0]]
        S_y = source_pos[indexes[1]]
        S_z = source_pos[indexes[2]]

        half_l = dd_l/2
        half_m = dd_m/2

        sin_phi_delta = (dl + half_l - S_x)/ (np.sqrt(np.square(
            dl+half_l-S_x) + np.square(dm-S_y) + np.square(dn-S_z)))
        test1 = (dl - half_l) <= S_x
        test2 = S_x <= (dl + half_l)
        k_phi = -1 if test1 and test2 else 1

        sin_phi = k_phi * (dl - half_l - S_x) / (np.sqrt(np.square(
            dl-half_l-S_x) + np.square(dm-S_y) + np.square(dn-S_z)))
        if (sin_phi_delta-sin_phi) < 1e-11:
            sin_phi *= -1

        plus  = np.arctan(np.abs((dm+half_m-S_y)/np.abs(S_z)))
        minus = np.arctan(np.abs((dm-half_m-S_y)/np.abs(S_z)))

        test1 = (dm - half_m) <= S_y
        test2 = S_y <= (dm + half_m)
        k_beta = -1 if test1 and test2 else 1

        beta = np.abs(plus-(k_beta*minus))
        distance_out[j] = np.sqrt(
            np.square(dl-S_x) + np.square(dm-S_y) + np.square(dn-S_z))

        if air_attenuation is not None:
            energy[j, :] = (np.abs(sin_phi_delta-sin_phi) ) * beta / (
                4*np.pi) * np.exp(
                -air_attenuation * distance_out[j])
        else:
            energy[j, :] = (np.abs(sin_phi_delta-sin_phi) ) * beta / (
                4*np.pi)

    return (energy, distance_out)

def _patch2receiver_energy_kang(
        patch_receiver_distance, patches_normal):
    receiver_factor = np.empty((
        patch_receiver_distance.shape[0],))
    for i in range(patch_receiver_distance.shape[0]):
        R = np.sqrt(np.sum((patch_receiver_distance[i, :]**2)))

        cos_xi = np.abs(np.sum(
            patches_normal[i, :]*np.abs(patch_receiver_distance[i, :]))) / R

        # Equation 20
        receiver_factor[i] = cos_xi / (np.pi * R**2)

    return receiver_factor

if numba is not None:
    patch2patch_ff_kang = numba.njit(parallel=True)(patch2patch_ff_kang)
    _source2patch_energy_kang = numba.njit(parallel=True)(
        _source2patch_energy_kang)
    _patch2receiver_energy_kang = numba.njit()(_patch2receiver_energy_kang)



import pytest
import numpy as np
import numpy.testing as npt
import sparrowpy as sp
import pyfar as pf


def test_init_default_single_wall():
    walls = sp.testing.shoebox_room_stub(1, 1, 1)
    walls = [walls[0]]
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        walls, 1)

    npt.assert_almost_equal(radiosity.patches_points.shape, (1, 4, 3))
    npt.assert_almost_equal(radiosity.patches_area.shape, (1,))
    npt.assert_almost_equal(radiosity.patches_center.shape, (1, 3))
    npt.assert_almost_equal(radiosity.patches_size.shape, (1, 3))
    npt.assert_almost_equal(radiosity.patches_normal.shape, (1, 3))


@pytest.mark.parametrize("apply_brdf", [True, False])
@pytest.mark.parametrize("apply_attenuation", [True, False])
def test_io(apply_brdf, apply_attenuation, tmpdir):
    walls = sp.testing.shoebox_room_stub(1, 1, 1)
    walls = walls[:2]
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        walls, 1)
    radiosity.write(tmpdir / "test.far")
    radiosity_out = sp.DirectionalRadiosityFast.from_read(
        tmpdir / "test.far")
    assert radiosity_out == radiosity

    # set brdf and air attenuation
    frequencies = [1000]
    if apply_brdf:
        radiosity.set_wall_brdf(
            np.arange(len(walls)),
            pf.FrequencyData(np.ones_like(frequencies), frequencies),
            pf.Coordinates(0, 0, 1, weights=1),
            pf.Coordinates(0, 0, 1, weights=1))
    if apply_attenuation:
        radiosity.set_air_attenuation(
            pf.FrequencyData(np.ones_like(frequencies), frequencies))
    if apply_attenuation or apply_brdf:
        radiosity.write(tmpdir / "test.far")
        radiosity_out = sp.DirectionalRadiosityFast.from_read(
            tmpdir / "test.far")
        assert radiosity_out == radiosity

    # check readwrite after baking
    radiosity.bake_geometry()
    radiosity.write(tmpdir / "test.far")
    radiosity_out = sp.DirectionalRadiosityFast.from_read(
        tmpdir / "test.far")
    assert radiosity_out == radiosity

    # check readwrite after init_source_energy
    radiosity.init_source_energy(pf.Coordinates(.5, .5, .5))
    radiosity.write(tmpdir / "test.far")
    radiosity_out = sp.DirectionalRadiosityFast.from_read(
        tmpdir / "test.far")
    assert radiosity_out == radiosity

    # check readwrite after init_source_energy
    radiosity.calculate_energy_exchange(343, 1/1000, 1, 3)
    radiosity.write(tmpdir / "test.far")
    radiosity_out = sp.DirectionalRadiosityFast.from_read(
        tmpdir / "test.far")
    assert radiosity_out == radiosity


def test_io_within_simulation(tmpdir):
    walls = sp.testing.shoebox_room_stub(1, 1, 1)
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        walls, 1)

    # set brdf and air attenuation
    brdf_sources = pf.Coordinates(0, 0, 1, weights=1)
    brdf_receivers = pf.Coordinates(0, 0, 1, weights=1)
    frequencies = np.array([1000])
    brdf = sp.brdf.create_from_scattering(
        brdf_sources,
        brdf_receivers,
        pf.FrequencyData(1, frequencies),
        pf.FrequencyData(.1, frequencies))

    # set directional scattering data
    radiosity.set_wall_brdf(
        np.arange(len(walls)), brdf, brdf_sources, brdf_receivers)
    radiosity.set_air_attenuation(
        pf.FrequencyData(
            np.zeros_like(brdf.frequencies),
            brdf.frequencies))
    radiosity.write(tmpdir / "test.far")
    radiosity_read = sp.DirectionalRadiosityFast.from_read(
        tmpdir / "test.far")

    # check readwrite after baking
    radiosity_read.bake_geometry()
    radiosity.bake_geometry()

    radiosity.write(tmpdir / "test.far")
    radiosity_read = sp.DirectionalRadiosityFast.from_read(
        tmpdir / "test.far")

    # check readwrite after init_source_energy
    radiosity.init_source_energy(pf.Coordinates(.5, .5, .5))
    radiosity_read.init_source_energy(pf.Coordinates(.5, .5, .5))

    radiosity.write(tmpdir / "test.far")
    radiosity_read = sp.DirectionalRadiosityFast.from_read(
        tmpdir / "test.far")

    # check readwrite after init_source_energy
    radiosity.calculate_energy_exchange(
        343, 1/1000, 1, 3, recalculate=True)
    radiosity_read.calculate_energy_exchange(
        343, 1/1000, 1, 3, recalculate=True)

    radiosity.write(tmpdir / "test.far")
    radiosity_read = sp.DirectionalRadiosityFast.from_read(
        tmpdir / "test.far")

    radiosity.collect_energy_receiver_mono(pf.Coordinates(.1, .5, .5))
    radiosity_read.collect_energy_receiver_mono(pf.Coordinates(.1, .5, .5))



import numpy as np
import sparrowpy.sound_object as so
import os
import numpy.testing as npt


def test_sound_source_init():
    """Test SoundSource initialization."""
    position = np.array([0, 0, 0])
    view = np.array([1, 0, 0])
    up = np.array([0, 0, 1])
    path = os.path.join(
        os.path.dirname(__file__), 'test_data',
        'Genelec8020_DAF_2016_1x1.v17.ms.sofa')
    directivity = so.DirectivityMS(path)
    sound_source = so.SoundSource(position, view, up, directivity)

    # check any direction
    assert isinstance(sound_source, so.SoundSource)
    assert isinstance(sound_source.directivity, so.DirectivityMS)
    res = sound_source.get_directivity(view, 1000)
    npt.assert_almost_equal((2009.3389892578125+0j), res)
    res = sound_source.get_directivity(up, 1000)
    npt.assert_almost_equal((3343.79980469+0j), res)

    # set different view and check if rotation works
    view = np.array([0, 1, 0])
    sound_source = so.SoundSource(position, view, up, directivity)
    res = sound_source.get_directivity(view, 1000)
    npt.assert_almost_equal((2009.3389892578125+0j), res)
    res = sound_source.get_directivity(up, 1000)
    npt.assert_almost_equal((3343.79980469+0j), res)


def test_sound_source_multiple_targets():
    """Test SoundSource multiple targets."""
    position = np.array([0, 0, 0])
    view = np.array([1, 0, 0])
    up = np.array([0, 0, 1])
    path = os.path.join(
        os.path.dirname(__file__), 'test_data',
        'Genelec8020_DAF_2016_1x1.v17.ms.sofa')
    directivity = so.DirectivityMS(path)
    sound_source = so.SoundSource(position, view, up, directivity)

    # check any direction
    assert isinstance(sound_source, so.SoundSource)
    assert isinstance(sound_source.directivity, so.DirectivityMS)
    res = sound_source.get_directivity(np.array((view, up)), 1000)
    npt.assert_almost_equal(res.shape, (2,))
    npt.assert_almost_equal((2009.3389892578125+0j), res[0])
    npt.assert_almost_equal((3343.79980469+0j), res[1])

    # set different view and check if rotation works
    view = np.array([0, 1, 0])
    sound_source = so.SoundSource(position, view, up, directivity)
    res = sound_source.get_directivity(np.array((view, up)), 1000)
    npt.assert_almost_equal((2009.3389892578125+0j), res[0])
    npt.assert_almost_equal((3343.79980469+0j), res[1])



"""methods for universal form factor calculation."""
import numpy as np
import sparrowpy.geometry as geom
from sparrowpy.form_factor import integration
try:
    import numba
    prange = numba.prange
except ImportError:
    numba = None
    prange = range

def patch2patch_ff_universal(patches_points: np.ndarray,
                             patches_normals: np.ndarray,
                             patches_areas: np.ndarray,
                             visible_patches:np.ndarray):
    """Calculate the form factors between patches (universal method).

    This method computes form factors between polygonal patches via numerical
    integration. Support is only guaranteed for convex polygons.
    Because this function relies on numpy, all patches must have
    the same number of sides.

    Parameters
    ----------
    patches_points : np.ndarray
        vertices of all patches of shape (n_patches, n_vertices, 3)
    patches_normals : np.ndarray
        normal vectors of all patches of shape (n_patches, 3)
    patches_areas : np.ndarray
        areas of all patches of shape (n_patches)
    visible_patches : np.ndarray
        index list of all visible patches combinations (n_combinations, 2)

    Returns
    -------
    form_factors : np.ndarray
        form factors between all patches of shape (n_patches, n_patches)
        note that just i_source < i_receiver are calculated since
        patches_areas[i] * ff[i, j] = patches_areas[j] * ff[j, i]

    """
    n_patches = patches_areas.shape[0]
    form_factors = np.zeros((n_patches, n_patches))

    for visID in prange(visible_patches.shape[0]):
        i = int(visible_patches[visID, 0])
        j = int(visible_patches[visID, 1])
        form_factors[i,j] = universal_form_factor(
                    patches_points[i], patches_normals[i], patches_areas[i],
                    patches_points[j], patches_normals[j])

    return form_factors

def universal_form_factor(source_pts: np.ndarray, source_normal: np.ndarray,
                     source_area: np.ndarray, receiver_pts: np.ndarray,
                     receiver_normal: np.ndarray,
                     ) -> float:
    """Return the form factor based on input patches geometry.

    Parameters
    ----------
    receiver_pts: np.ndarray
        receiver patch vertex coordinates (n_vertices,3)

    receiver_normal: np.ndarray
        receiver patch normal (3,)

    receiver_area: float
        receiver patch area

    source_pts: np.ndarray
        source patch vertex coordinates (n_vertices,3)

    source_normal: np.ndarray
        source patch normal (3,)

    source_area: float
        source patch area

    Returns
    -------
    form_factor: float
        form factor

    """
    if geom._coincidence_check(receiver_pts, source_pts):
        form_factor = integration.nusselt_integration(
                    patch_i=source_pts, patch_i_normal=source_normal,
                    patch_j=receiver_pts, patch_j_normal=receiver_normal,
                    nsamples=64)
    else:
        form_factor = integration.stokes_integration(patch_i=source_pts,
                                             patch_j=receiver_pts,
                                             patch_i_area=source_area)

    return form_factor

def _source2patch_energy_universal(
        source_position: np.ndarray, patches_center: np.ndarray,
        patches_points: np.ndarray, source_visibility: np.ndarray,
        air_attenuation:np.ndarray, n_bins:float):
    """Calculate the initial energy from the source.

    Parameters
    ----------
    source_position : np.ndarray
        source position of shape (3,)
    patches_center : np.ndarray
        center of all patches of shape (n_patches, 3)
    patches_points : np.ndarray
        vertices of all patches of shape (n_patches, n_points, 3)
    source_visibility : np.ndarray
        visibility condition between source and patches (n_patches)
    air_attenuation : np.ndarray
        air attenuation factor in Np/m (n_bins,)
    n_bins : float
        number of frequency bins.

    Returns
    -------
    energy : np.ndarray
        energy of all patches of shape (n_patches)
    distance : np.ndarray
        corresponding distance of all patches of shape (n_patches)

    """
    n_patches = patches_center.shape[0]
    energy = np.zeros((n_patches, n_bins))
    distance_out = np.zeros((n_patches, ))
    for j in prange(n_patches):
        if source_visibility[j]:
            source_pos = source_position.copy()
            receiver_pos = patches_center[j, :].copy()
            receiver_pts = patches_points[j, :, :].copy()

            distance_out[j] = np.linalg.norm(source_pos-receiver_pos)

            if air_attenuation is not None:
                energy[j,:] = np.exp(
                    -air_attenuation*distance_out[j])*integration.pt_solution(
                        point=source_pos, patch_points=receiver_pts,
                        mode="source")
            else:
                energy[j,:] = integration.pt_solution(
                    point=source_pos, patch_points=receiver_pts, mode="source")

    return (energy, distance_out)

def _patch2receiver_energy_universal(
        receiver_pos, patches_points, receiver_visibility):

    receiver_factor = np.zeros((patches_points.shape[0]))


    for i in prange(patches_points.shape[0]):
        if receiver_visibility[i]:
            receiver_factor[i] = integration.pt_solution(point=receiver_pos,
                            patch_points=patches_points[i,:], mode="receiver")

    return receiver_factor


if numba is not None:
    patch2patch_ff_universal = numba.njit(parallel=True)(
        patch2patch_ff_universal)
    universal_form_factor = numba.njit()(universal_form_factor)
    _source2patch_energy_universal = numba.njit(parallel=True)(
        _source2patch_energy_universal)
    _patch2receiver_energy_universal = numba.njit(parallel=True)(
        _patch2receiver_energy_universal)
